/**
 * Copyright © 2021 Maxime Friess <M4x1me@pm.me>
 *
 * This file is part of Seb-BOT.
 *
 * Seb-BOT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Seb-BOT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Seb-BOT.  If not, see <https://www.gnu.org/licenses/>.
 */

import { GetPeopleRequest, GetPersonalAccountsRequest, GetPersonalTransactionsRequest } from '@amicale-core/seb-ts';
import { APIApplicationCommandOption, ApplicationCommandOptionType } from 'discord-api-types/v9';
import { ButtonInteraction, CommandInteraction, MessageActionRow, MessageButton, MessageEmbed, WebhookEditMessageOptions } from 'discord.js';
import { Command } from '../base/Command';
import { Bot } from '../Bot';
import { SoftConfig } from '../config/SoftConfig';
import { I18n } from '../utils/I18n';

export class AccountCommand extends Command {
    private sub: { [index: string]: (interaction: CommandInteraction) => Promise<void> };

    constructor() {
        super();
        this.sub = {
            balance: this.exec_balance.bind(this),
            history: this.exec_history.bind(this),
        };

        Bot.registerButton('acc_history', this.historyButton.bind(this));
    }

    getName() {
        return 'account';
    }

    getOptions(): APIApplicationCommandOption[] {
        return [
            {
                type: ApplicationCommandOptionType.Subcommand,
                ...I18n.argumentI18n(this, 'balance')
            },
            {
                type: ApplicationCommandOptionType.Subcommand,
                ...I18n.argumentI18n(this, 'history')
            },
        ];
    }

    async get_account(interaction: CommandInteraction | ButtonInteraction) {
        const people_req = new GetPeopleRequest(Bot.getRest(), { query: { filter: { discord_id: interaction.user.id } } });
        const people = (await people_req.make()).data;

        if (people.length <= 0) {
            await interaction.editReply({ content: I18n.getI18n('error.notregistered', interaction.locale) });
            return;
        }

        const accounts_req = new GetPersonalAccountsRequest(Bot.getRest(), { query: { filter: { person_id: people[0].id } } });
        const accounts = (await accounts_req.make()).data;

        if (accounts.length <= 0) {
            await interaction.editReply({ content: I18n.getI18n("command.account.error.noaccount", interaction.locale) });
            return;
        }

        return { d: people[0], a_d: accounts[0] };
    }

    private async historyButton(interaction: ButtonInteraction) {
        await interaction.deferUpdate();
        const page = parseInt(interaction.customId.split(',')[1]);
        await interaction.editReply(await this.history_message(interaction, page));
    }

    private async history_message(interaction: CommandInteraction | ButtonInteraction, page = 1): Promise<WebhookEditMessageOptions> {
        const acc = await this.get_account(interaction);
        if (acc === undefined) return {};
        const { d, a_d } = acc;

        const transactions_req = new GetPersonalTransactionsRequest(Bot.getRest(), {
            query: { filter: { personal_account_id: a_d.id }, per_page: 10, page: page, order_by: 'created_at', order_sort: 'DESC' },
        });
        const transactions = await transactions_req.make();

        if (transactions.data.length <= 0) {
            await interaction.editReply({ content: I18n.getI18n('command.account.error.notransaction', interaction.locale) });
            return {};
        }

        const spaces = '\n'.repeat(10 - transactions.data.length) + (transactions.data.length !== 10 ? '🚎' : '');

        const embed = new MessageEmbed()
            .setColor(SoftConfig.get('bot.color', '#fb963a'))
            .addField(I18n.getI18n('command.account.reply.history.id', interaction.locale), transactions.data.map((e) => '#' + e.id).join('\n') + spaces, true)
            .addField(I18n.getI18n('command.account.reply.history.date', interaction.locale), transactions.data.map((e) => new Date(e.created_at ?? 0).toLocaleString('fr-FR')).join('\n'), true)
            .addField(
                I18n.getI18n('command.account.reply.history.amount', interaction.locale),
                transactions.data
                    .map((e) => Number(e.amount).toLocaleString('fr-FR', { currency: 'EUR', currencyDisplay: 'symbol', style: 'currency' }))
                    .join('\n'),
                true
            );

        const row = new MessageActionRow().addComponents(
            new MessageButton()
                .setCustomId(`acc_history,${transactions.metadata.current_page - 1}`)
                .setLabel('<')
                .setDisabled(transactions.metadata.current_page == 1)
                .setStyle('PRIMARY'),
            new MessageButton()
                .setCustomId(`acc_history_page`)
                .setLabel(`${transactions.metadata.current_page} / ${transactions.metadata.last_page}`)
                .setDisabled(true)
                .setStyle('SECONDARY'),
            new MessageButton()
                .setCustomId(`acc_history,${transactions.metadata.current_page + 1}`)
                .setLabel('>')
                .setDisabled(transactions.metadata.current_page == transactions.metadata.last_page)
                .setStyle('PRIMARY')
        );

        return {
            embeds: [embed],
            components: [row],
            content: I18n.formatI18n('command.account.reply.history', interaction.locale, {
                fullname: d.fullname
            })
        };
    }

    async exec_history(interaction: CommandInteraction) {
        await interaction.deferReply({ ephemeral: true });
        await interaction.editReply(await this.history_message(interaction));
    }

    async exec_balance(interaction: CommandInteraction) {
        await interaction.deferReply({ ephemeral: true });
        const acc = await this.get_account(interaction);
        if (acc === undefined) return;
        const { d, a_d } = acc;
        interaction.editReply({
            content: I18n.formatI18n('command.account.reply.balance', interaction.locale, {
                fullname: d.fullname,
                balance: Number(a_d.balance).toLocaleString('fr-FR', {
                    currency: 'EUR',
                    currencyDisplay: 'symbol',
                    style: 'currency',
                })
            })
        });
    }

    async execute(interaction: CommandInteraction) {
        await this.sub[interaction.options.getSubcommand()](interaction);
    }
}
