/**
 * Copyright © 2021 Maxime Friess <M4x1me@pm.me>
 *
 * This file is part of Seb-BOT.
 *
 * Seb-BOT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Seb-BOT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Seb-BOT.  If not, see <https://www.gnu.org/licenses/>.
 */

import { APIApplicationCommandOption, ApplicationCommandOptionType } from 'discord-api-types/v9';
import { ButtonInteraction, CommandInteraction, InteractionReplyOptions, MessageActionRow, MessageButton, MessageEmbed, MessageSelectMenu, SelectMenuInteraction } from 'discord.js';
import moment from 'moment';
import { CrousApi } from '../api/CrousApi';
import { Command } from '../base/Command';
import { Bot } from '../Bot';
import { SoftConfig } from '../config/SoftConfig';
import { Cache } from '../utils/Cache';
import { I18n } from '../utils/I18n';
import { RU } from '../utils/RU';
import { StringUtils } from '../utils/String';

const DATE_FORMAT = 'DD/MM/YYYY';
const DEFAULT_RU = "25,1392";
type FilterValues = 'ETU' | 'PER' | 'ALL';

export class MenuCommand extends Command {
    constructor() {
        super();
        Bot.registerButton('menu_date', this.buttonDate.bind(this));
        Bot.registerSelect('menu_switch', this.selectRestaurant.bind(this))
    }

    getName() {
        return 'menu';
    }

    getOptions(): APIApplicationCommandOption[] {
        return [
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "group"),
                choices: [
                    { ...I18n.choiceI18n(this, "group", "professor"), value: "PERS" },
                    { ...I18n.choiceI18n(this, "group", "student"), value: "ETU" },
                    { ...I18n.choiceI18n(this, "group", "all"), value: "ALL" },
                ],
                required: false,
            },
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "date"),
                required: false,
            },
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "restaurant"),
                choices: RU.getRUArguments(),
                required: false,
            }
        ];
    }

    async buttonDate(interaction: ButtonInteraction) {
        const [, d, f, re, ra] = interaction.customId.split(',');
        const date = moment(d, DATE_FORMAT).toDate();
        const filter = f as FilterValues;

        await interaction.deferUpdate();
        await interaction.editReply(await this.getMessageContent(date, filter, parseInt(re), parseInt(ra), interaction.locale));
    }

    async selectRestaurant(interaction: SelectMenuInteraction) {
        const [, d, f] = interaction.customId.split(",");
        const [region, restaurant] = interaction.values[0].split(",").map(x => parseInt(x));
        const date = moment(d, DATE_FORMAT).toDate();
        const filter = f as FilterValues;

        await interaction.deferUpdate();
        await interaction.editReply(await this.getMessageContent(date, filter, region, restaurant, interaction.locale));
    }

    async execute(interaction: CommandInteraction) {
        const date = interaction.options.getString('date') === null ? new Date() : moment(interaction.options.getString('date'), DATE_FORMAT).toDate();
        let filter = (interaction.options.getString('groupe') ?? 'ETU') as FilterValues;
        const [region, restaurant] = (interaction.options.getString('restaurant') ?? DEFAULT_RU).split(',').map(x => parseInt(x));

        if (isNaN(date.getTime())) {
            await interaction.reply({ content: I18n.getI18n("bot.error.date", interaction.locale), ephemeral: true });
            return;
        }

        await interaction.deferReply();
        await interaction.editReply(await this.getMessageContent(date, filter, region, restaurant, interaction.locale));
    }

    async getMessageContent(date: Date, filter: FilterValues, region: number, restaurant: number, lang: string): Promise<InteractionReplyOptions> {
        const menu = await CrousApi.menu(region, restaurant);

        const rows = [new MessageActionRow().addComponents(
            new MessageSelectMenu().setCustomId(`menu_switch,${moment(date).format(DATE_FORMAT)},${filter}`).setOptions(RU.getRUSelect(region, restaurant))
        ), new MessageActionRow().addComponents(
            new MessageButton().setCustomId(`menu_date,${moment(date).subtract(1, 'd').format(DATE_FORMAT)},${filter},${region},${restaurant}`).setLabel(I18n.getI18n("command.menu.reply.day.previous", lang)).setStyle('PRIMARY'),
            new MessageButton().setCustomId(`menu_date,${moment(date).add(1, 'd').format(DATE_FORMAT)},${filter},${region},${restaurant}`).setLabel(I18n.getI18n("command.menu.reply.day.next", lang)).setStyle('PRIMARY')
        )];

        for (const m of menu) {
            if (moment(m.date).isSame(date, 'day')) {
                for (const meal of m.meal) {
                    if (meal.name == 'midi') {
                        const embed = new MessageEmbed().setColor(SoftConfig.get('bot.color', '#fb963a')).setTimestamp(Cache.modified_date('crous.menu'));
                        for (const cat of meal.foodcategory.sort(this.__sortfunc)) {
                            // Filter is only needed for restaurant 1392 (Illkirch)
                            // TODO: Clean this a bit.
                            if (cat.name.includes(filter) || restaurant !== 1392)
                                embed.addField(
                                    StringUtils.capitalize(cat.name),
                                    cat.dishes.filter((x: any) => x.name !== "").map((x: { name: string }) => '> ' + StringUtils.capitalize(x.name)).join('\n'),
                                    false
                                );
                        }
                        return { embeds: [embed], content: I18n.formatI18n("command.menu.reply.title", lang, { filter, date: moment(date).format(DATE_FORMAT) }), components: rows };
                    }
                }
                return { content: I18n.formatI18n("command.menu.reply.error", lang, { filter, date: moment(date).format(DATE_FORMAT) }), components: rows, embeds: [] };
            }
        }
        return { content: I18n.formatI18n("command.menu.reply.error", lang, { filter, date: moment(date).format(DATE_FORMAT) }), components: rows, embeds: [] };
    }

    __sortfunc(a: { name: string }, b: { name: string }) {
        if (a.name.includes('PER') && b.name.includes('ETU')) {
            return -1;
        } else if (a.name.includes('ETU') && b.name.includes('PER')) {
            return 1;
        } else {
            if (a.name.includes('ENTREES') || b.name.includes('DESSERTS')) {
                return -1;
            } else if (a.name.includes('DESSERTS') || b.name.includes('ENTREES')) {
                return 1;
            } else {
                return a.name.localeCompare(b.name);
            }
        }
    }
}
