/**
 * Copyright © 2021 Maxime Friess <M4x1me@pm.me>
 *
 * This file is part of Seb-BOT.
 *
 * Seb-BOT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Seb-BOT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Seb-BOT.  If not, see <https://www.gnu.org/licenses/>.
 */

import { GetPeopleRequest, MissingFieldsException, StorePersonRequest } from '@amicale-core/seb-ts';
import { APIApplicationCommandOption, ApplicationCommandOptionType } from 'discord-api-types/v9';
import { ButtonInteraction, CommandInteraction, MessageActionRow, MessageButton } from 'discord.js';
import { Command } from '../base/Command';
import { Bot } from '../Bot';
import { SoftConfig } from '../config/SoftConfig';
import { I18n } from '../utils/I18n';

export class RegisterCommand extends Command {

    constructor() {
        super();
        Bot.registerButton('policy_agree', this.acceptButton.bind(this));
    }

    getName() {
        return 'register';
    }

    getOptions(): APIApplicationCommandOption[] {
        return [
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "firstname"),
                required: true,
            },
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "lastname"),
                required: true,
            },
        ];
    }

    async acceptButton(interaction: ButtonInteraction) {
        const [, firstname, lastname] = interaction.customId.split(",");
        await interaction.deferUpdate();

        const person_req = new StorePersonRequest(Bot.getRest(), {
            discord_id: interaction.user.id,
            firstname,
            lastname,
        });

        try {
            const person = (await person_req.make()).data;

            let message = '';
            message += I18n.formatI18n("command.register.reply.welcome1", interaction.locale, person) + "\n";
            message += I18n.getI18n("command.register.reply.welcome2", interaction.locale);

            interaction.editReply({ content: message, components: [] });
        } catch (e) {
            if (e instanceof MissingFieldsException) {
                interaction.editReply({
                    content: I18n.getI18n("command.register.reply.error.already", interaction.locale),
                    components: [],
                });
            } else {
                interaction.editReply({ content: I18n.getI18n("command.register.reply.error.conn", interaction.locale), components: [] });
            }
        }
    }

    async execute(interaction: CommandInteraction) {
        await interaction.deferReply({ ephemeral: true });
        const people_req = new GetPeopleRequest(Bot.getRest(), { query: { filter: { discord_id: interaction.user.id } } });
        const people = (await people_req.make()).data;

        if (people.length > 0) {
            await interaction.editReply({ content: I18n.getI18n("command.register.reply.error.already", interaction.locale) });
            return;
        }

        let firstname = interaction.options.getString('firstname');
        firstname = '' + firstname?.charAt(0).toUpperCase() + firstname?.slice(1).toLowerCase();
        firstname = firstname.replaceAll(",", "");
        const lastname = ('' + interaction.options.getString('lastname')?.toUpperCase()).replaceAll(",", "");

        const row = new MessageActionRow()
            .addComponents(
                new MessageButton()
                    .setLabel(I18n.getI18n("command.register.reply.privacy", interaction.locale))
                    .setURL(SoftConfig.get('seb.url', '') + 'doc/privacy-policy.pdf')
                    .setStyle('LINK')
            )
            .addComponents(new MessageButton().setCustomId(`policy_agree,${firstname},${lastname}`).setLabel('Accepter').setStyle('SUCCESS'));

        await interaction.editReply({
            content: I18n.getI18n("command.register.reply.agree", interaction.locale),
            components: [row],
        });
    }
}
