/**
 * Copyright © 2021 Maxime Friess <M4x1me@pm.me>
 *
 * This file is part of Seb-BOT.
 *
 * Seb-BOT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Seb-BOT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Seb-BOT.  If not, see <https://www.gnu.org/licenses/>.
 */

import { APIApplicationCommandOption, ApplicationCommandOptionType } from 'discord-api-types/v9';
import { CommandInteraction, MessageActionRow, MessageAttachment, MessageSelectMenu, SelectMenuInteraction, WebhookEditMessageOptions } from 'discord.js';
import moment from 'moment';
import { AdeApi } from '../api/AdeApi';
import { Command } from '../base/Command';
import { Bot } from '../Bot';
import { I18n } from '../utils/I18n';
import MapManager from '../utils/MapManager';

const DATE_FORMAT = 'DD/MM/YYYY hh:mm';

export class RoomsCommands extends Command {
    constructor() {
        super();
        Bot.registerSelect('rooms_floor', this.selectFloor.bind(this));
    }

    getName() {
        return 'rooms';
    }

    getOptions(): APIApplicationCommandOption[] {
        return [
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "building"),
                choices: MapManager.getMaps(),
                required: true,
            },
            {
                type: ApplicationCommandOptionType.String,
                ...I18n.argumentI18n(this, "date"),
                required: false,
            },
        ];
    }

    __parseDate(date: string | null) {
        if (date === null) {
            return new Date();
        } else {
            return moment(date, DATE_FORMAT).toDate();
        }
    }

    async __getMessageData(bat: string, date: Date, lang:string): Promise<WebhookEditMessageOptions> {
        const rooms = await AdeApi.used_rooms_at(date);

        const data = MapManager.genMap(bat, rooms);

        const row = new MessageActionRow().addComponents(
            new MessageSelectMenu().setCustomId(`rooms_floor,${moment(date).format(DATE_FORMAT)}`).setPlaceholder(I18n.getI18n("command.rooms.reply.building", lang)).addOptions(MapManager.getSelectMaps())
        );

        return { content: I18n.formatI18n("command.rooms.reply.title", lang, {bat, d: moment(date).format(DATE_FORMAT)}), files: [new MessageAttachment(data, 'rooms.png')], components: [row] };
    }

    async selectFloor(interaction: SelectMenuInteraction) {
        await interaction.deferUpdate();
        const date = this.__parseDate(interaction.customId.split(",")[1]);
        const bat = interaction.values[0];

        const msg = await this.__getMessageData(bat, date ?? new Date(), interaction.locale);
        await interaction.editReply(msg);
    }

    async execute(interaction: CommandInteraction) {
        const date = this.__parseDate(interaction.options.getString('date'));
        const bat = interaction.options.getString('building') ?? '';
        if (date === null) {
            await interaction.reply({ content: I18n.getI18n("bot.error.date", interaction.locale), ephemeral: true });
            return;
        }

        await interaction.deferReply();

        const msg = await this.__getMessageData(bat, date, interaction.locale);
        await interaction.editReply(msg);
    }
}
