import canvas, { Image } from 'canvas';
import fs from 'fs';
import path from 'path';
import { Log, Logger } from './Logger';

type MapData = {
    name: string;
    file: string;
    image?: Image;
    rooms: {
        name: string;
        x: number;
        y: number;
        w: number;
        h: number;
    }[];
}[];

class MapManager {
    private logger: Log;
    private maps: MapData;
    private maps_list: string[];

    constructor() {
        this.logger = Logger.getLogger('MapManager');
        this.maps = [];
        this.maps_list = [];
    }

    async load() {
        try {
            this.logger.info('Loading maps...');
            const raw = fs.readFileSync(path.join('resources', 'map', 'map.json'));
            this.maps = JSON.parse(raw.toString()).data;
        } catch (e: any) {
            this.logger.fatal('Error while loading maps!', e, 5);
        }
        this.maps_list = [];

        for (const map_id in this.maps) {
            this.maps_list.push(this.maps[map_id].name);
            try {
                const p = path.join('resources', 'map', this.maps[map_id].file);
                this.logger.info('Loading ' + p + '...');
                this.maps[map_id].image = await canvas.loadImage(p);
            } catch (e: any) {
                this.logger.fatal('Error while loading maps!', e, 5);
            }
        }

        this.logger.info('Loaded maps.');
    }

    getMaps() {
        const list = [];

        for (const d of this.maps_list) {
            list.push({
                name: d,
                value: d,
            });
        }

        return list;
    }

    getSelectMaps() {
        const list = [];

        for (const d of this.maps_list) {
            list.push({
                label: d,
                value: d,
            });
        }

        return list;
    }

    genMap(name: string, used_rooms: string[]): Buffer {
        let data = null;

        for (const map_id in this.maps) {
            if (this.maps[map_id].name === name) {
                data = this.maps[map_id];
            }
        }

        if (data === null || data.image === undefined) {
            const cnv = canvas.createCanvas(1, 1);
            return cnv.toBuffer();
        }

        const cnv = canvas.createCanvas(data.image.width, data.image.height);
        const ctx = cnv.getContext('2d');
        ctx.fillStyle = 'white';
        ctx.fillRect(0, 0, data.image.width, data.image.height);

        for (const room_id in data.rooms) {
            if (used_rooms.includes(data.rooms[room_id].name)) {
                ctx.fillStyle = '#ff6060';
            } else {
                ctx.fillStyle = '#adff80';
            }
            ctx.fillRect(data.rooms[room_id].x, data.rooms[room_id].y, data.rooms[room_id].w, data.rooms[room_id].h);
        }

        ctx.drawImage(data.image, 0, 0, data.image.width, data.image.height);

        return cnv.toBuffer();
    }
}

export default new MapManager();
